from torch.utils.data import Dataset
from torchvision import transforms
from PIL import Image
import torch
import random
import glob
import json
import pandas as pd
from sg_diffuser.types import SceneGraph
#from itertools import chain


class CustomDataset(Dataset):

    def __init__(self, image_dir, data_dir, transform, drop_rate=0):
        super().__init__()
        self.images = sorted(
            glob.glob(f'{image_dir}/**/*.jpg', recursive=True), key=lambda x: x.split('/')[-1].lstrip('0')
        )
        self.scene_graphs = sorted(
            glob.glob(f'{data_dir}/**/*.json', recursive=True), key=lambda x: x.split('/')[-1].lstrip('0')
        )

        assert len(self.images) == len(self.scene_graphs)
        assert all(img.split('/')[-1].strip('.jpg').lstrip('0') == sg.split('/')[-1].strip('.json').lstrip('0')
                   for img, sg in zip(self.images, self.scene_graphs))

        self.transform = transform
        self.drop_rate = drop_rate

    def __getitem__(self, idx):
        raw_image = Image.open(self.images[idx]).convert("RGB")
        image = self.transform(raw_image)

        with open(self.scene_graphs[idx], 'r') as f:
            json_data = json.load(f)

        nodes = {**json_data['rel_dict'], **json_data['EL_dict']}
        sorted_nodes = sorted(nodes, key=nodes.get)

        adj_list = json_data['adj_list']
        #edge_index = torch.tensor(adj_list).t().contiguous()
        edge_index = (
            torch.tensor(adj_list).t().contiguous() if len(adj_list) > 0 else torch.tensor([[], []], dtype=torch.int64)
        )

        if self.drop_rate > 0 and random.random() < self.drop_rate:
            sorted_nodes = [''] * len(sorted_nodes)
            edge_index = torch.tensor([[], []], dtype=torch.int64)

        return {
            'image': image,
            'scene_graph': SceneGraph(
                nodes=sorted_nodes,
                adj_list=edge_index,
                node_count=len(sorted_nodes),
                edge_count=edge_index.shape[1]
            )
        }

    def __len__(self):
        return len(self.images)


def collate_fn(data):
    images = torch.stack([example['image'] for example in data])
    nodes = sum([example['scene_graph'].nodes for example in data], [])
    node_counts = [example['scene_graph'].node_count for example in data]
    edge_counts = [example['scene_graph'].edge_count for example in data]

    node_cumsum = torch.cumsum(torch.tensor([0] + node_counts[:-1]), dim=0)
    #offsets = torch.tensor(list(chain.from_iterable([[v.item()]*c for v, c in zip(node_cumsum, edge_counts)])))
    offsets = torch.tensor(sum([[v.item()]*c for v, c in zip(node_cumsum, edge_counts)], []))
    edge_index = torch.cat([example['scene_graph'].adj_list for example in data], dim=1) + offsets

    return {
        'images': images,
        'scene_graphs': SceneGraph(
            nodes=nodes,
            adj_list=edge_index,
            node_count=len(nodes),
            edge_count=edge_index.shape[1]
        ),
        'node_counts': node_counts,
        'edge_counts': edge_counts
    }


def get_transforms_rand_crop(size=512):
    return transforms.Compose([
        transforms.Resize(size, interpolation=transforms.InterpolationMode.BILINEAR),
        transforms.RandomCrop(size),
        transforms.ToTensor(),
        transforms.Normalize([0.5], [0.5]),
    ])


def get_transforms_center_crop(size=512):
    return transforms.Compose([
        transforms.Resize(size, interpolation=transforms.InterpolationMode.BILINEAR),
        transforms.CenterCrop(size),
        transforms.ToTensor(),
        transforms.Normalize([0.5], [0.5]),
    ])


def get_transforms_no_crop(size=512):
    return transforms.Compose([
        transforms.Resize(size, interpolation=transforms.InterpolationMode.BILINEAR),
        transforms.ToTensor(),
        transforms.Normalize([0.5], [0.5]),
    ])


class CaptionDataset(Dataset):

    def __init__(self, data_path):
        super().__init__()
        self.data = pd.read_csv(data_path, usecols=['key', 'vlm_caption'], dtype=str)

    def __getitem__(self, idx):
        key = str(self.data.iloc[idx]['key'])
        caption = self.data.iloc[idx]['vlm_caption']
        return {'key': key, 'vlm_caption': caption}

    def __len__(self):
        return len(self.data)
