from datasets import load_dataset
from collections import Counter
from PIL import Image
from io import BytesIO
from tqdm import tqdm
import json
import urllib.request
import pandas
import os
import shutil
import numpy as np
import argparse


def main(args):
    # Load the dataset (~16.9M)
    ds = load_dataset("tomg-group-umd/pixelprose")

    # Filter the dataset (~2.4M)
    ds_f = ds['train'].filter(
        lambda x: x['original_width'] >= 512
                  and x['original_height'] >= 512
                  and x['watermark_class_id'] == 1  # images without watermark (and w/o text)
                  and x['watermark_class_score'] >= 0.8
                  and x['aesthetic_score'] > 4.5
                  and (max(x['original_width'], x['original_height']) / min(x['original_width'], x['original_height'])) <= 1.5  # aspect ratio 3:2/2:3 (or 5:3/3:5)
                  and len(x['vlm_caption'].split('\n')) <= 3  # max 2 '\n'
                  and all(ord(c) >= 0 and ord(c) <= 127 for c in x['vlm_caption'])
                  and not any(c in x['vlm_caption'] for c in r'\|"$%&/()=^*+[]#@<>')
                  and not 'slideplayer' in x['url']
                  and len(x['vlm_caption'].split(',')) <= 80  # max 79 ','
    )

    print(f"Total number of images: {ds_f.num_rows}, subset size: {args.set_size}")

    # Download the subset
    checkpointing_steps = args.checkpointing_steps
    resume_from_checkpoint = args.resume_from_checkpoint
    resume_from_rand = args.resume_from_rand
    last_ckp = args.last_ckp
    rand_last_ckp_path = args.rand_last_ckp_path
    restart = args.restart

    if restart:
        shutil.rmtree('./dataset/PixelProse_MVS', ignore_errors=True)

    output_path = './dataset/PixelProse_MVS'
    imgs_output_path = f'{output_path}/images'
    csv_output_path = f'{output_path}/csv'
    ckp_output_path = f'{output_path}/checkpoints'
    stats_output_path = f'{output_path}/stats'
    os.makedirs(output_path, exist_ok=True)
    os.makedirs(imgs_output_path, exist_ok=True)
    os.makedirs(csv_output_path, exist_ok=True)
    os.makedirs(ckp_output_path, exist_ok=True)
    os.makedirs(stats_output_path, exist_ok=True)

    headers = {
        "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:72.0) Gecko/20100101 Firefox/72.0"
    }

    if resume_from_checkpoint:
        print(f'Resuming from checkpoint {last_ckp}..')
        with open(f'{ckp_output_path}/checkpoint-{last_ckp}.json', 'r') as f:
            ckp = json.load(f)
        num_imgs = int(ckp['num_imgs'])
        set_size = int(ckp['set_size'])
        img_indexes = list(ckp['img_indexes'])
        unavailable_img_indexes_set = set(ckp['unavailable_img_indexes_set'])
        counter = Counter(ckp['counter'])
        deltas = list(ckp['deltas'])
        restart_vocabulary = int(ckp['restart_vocabulary'])
        restart_vocabulary_indexes = list(ckp['restart_vocabulary_indexes'])
        downloads_failed = int(ckp['downloads_failed'])
        words_list = np.array([set(wl) for wl in ckp['words_list']])
    elif resume_from_rand:
        print(f'Resuming from mvs {rand_last_ckp_path}..')
        with open(rand_last_ckp_path, 'r') as f:
            ckp = json.load(f)
        num_imgs = int(ckp['num_imgs'])
        set_size = int(ckp['set_size']) + args.set_size
        img_indexes = list(ckp['img_indexes'])
        unavailable_img_indexes_set = set(ckp['unavailable_img_indexes_set'])
        counter = Counter()
        deltas = []
        restart_vocabulary = 0
        restart_vocabulary_indexes = []
        downloads_failed = int(ckp['downloads_failed'])
        # words_list = np.array([set(wl) for wl in ckp['words_list']])
        words_list = np.array(
            [{w.lower().strip('_-.,;:!?\'\"\“\”') for w in prompt['vlm_caption'].split()} for prompt in tqdm(ds_f)]
            # [{w.lower().strip('_-.,;:!?\'\"\“\”') for w in ds_f[i]['vlm_caption'].split()} for i in tqdm(range(800000))]
        )
    else:
        num_imgs = 0
        set_size = args.set_size
        img_indexes = []
        unavailable_img_indexes_set = set()
        counter = Counter()
        deltas = []
        restart_vocabulary = 0
        restart_vocabulary_indexes = []
        downloads_failed = 0
        words_list = np.array(
            [{w.lower().strip('_-.,;:!?\'\"\“\”') for w in prompt['vlm_caption'].split()} for prompt in tqdm(ds_f)]
            #[{w.lower().strip('_-.,;:!?\'\"\“\”') for w in ds_f[i]['vlm_caption'].split()} for i in tqdm(range(800000))]
        )

    state = {
        'num_imgs': num_imgs,
        'set_size': set_size,
        'img_indexes': img_indexes,
        'unavailable_img_indexes_set': list(unavailable_img_indexes_set),
        'counter': dict(counter),
        'deltas': deltas,
        'restart_vocabulary': restart_vocabulary,
        'restart_vocabulary_indexes': restart_vocabulary_indexes,
        'downloads_failed': downloads_failed,
        'words_list': [list(ws) for ws in words_list],
    }
    with open(f'{ckp_output_path}/checkpoint-0.json', 'w', encoding='utf-8') as f:
        json.dump(state, f, ensure_ascii=False)

    len_func = np.vectorize(lambda x: len(x))

    while num_imgs < set_size:

        tmp_delta_list = len_func(words_list - set(counter.keys()))
        tmp_delta_list[list(unavailable_img_indexes_set.union(img_indexes))] = 0

        delta = int(tmp_delta_list.max())
        index = int(tmp_delta_list.argmax())
        print(f'{num_imgs}: {set_size} | {delta}, {index}')

        if delta > 0:
            try:
                # print(ds_f[index]['url'])
                request_ = urllib.request.Request(ds_f[index]['url'], None, headers)  # the assembled request
                response = urllib.request.urlopen(request_, timeout=8)  # store the response
                #if response.info().get_content_maintype() != 'image':
                #    print('not image, skipping..')
                #    raise ValueError('Invalid content')
            except:
                unavailable_img_indexes_set.add(index)
                downloads_failed += 1
                continue

            imgs_curr_dir = f'{imgs_output_path}/{str(num_imgs // checkpointing_steps).zfill(8)}'
            os.makedirs(imgs_curr_dir, exist_ok=True)

            ## create a new file and write the image
            ##with open(f"{imgs_curr_dir}/{ds_f[index]['key']}.jpg", 'wb') as f:
            ##    f.write(response.read())

            # check if image is corrupted
            try:
                #_ = Image.open(f"{imgs_curr_dir}/{ds_f[index]['key']}.jpg")
                img = Image.open(BytesIO(response.read()))
                img.save(f"{imgs_curr_dir}/{ds_f[index]['key']}.jpg")
            except:
                unavailable_img_indexes_set.add(index)
                downloads_failed += 1
                print('failed image open, skipping..')
                continue

            img_indexes.append(index)
            counter.update(Counter([w.lower().strip('_-.,;:!?\'\"\“\”') for w in ds_f[index]['vlm_caption'].split()]))
            deltas.append(delta)
            num_imgs += 1
        else:
            print(f'Vocabulary saturated after {num_imgs} images.')
            with open(f'{stats_output_path}/vocabulary-restart-{restart_vocabulary}.json', 'w', encoding='utf-8') as f:
                json.dump(dict(counter), f, ensure_ascii=False)
            counter = Counter()
            restart_vocabulary_indexes.append(num_imgs)
            restart_vocabulary += 1

        if (num_imgs % checkpointing_steps) == 0:
            state = {
                'num_imgs': num_imgs,
                'set_size': set_size,
                'img_indexes': img_indexes,
                'unavailable_img_indexes_set': list(unavailable_img_indexes_set),
                'counter': dict(counter),
                'deltas': deltas,
                'restart_vocabulary': restart_vocabulary,
                'restart_vocabulary_indexes': restart_vocabulary_indexes,
                'downloads_failed': downloads_failed,
                'words_list': [list(ws) for ws in words_list],
            }
            with open(f'{ckp_output_path}/checkpoint-{num_imgs // checkpointing_steps}.json', 'w', encoding='utf-8') as f:
                json.dump(state, f, ensure_ascii=False)
            # save dataset split csv
            split = pandas.DataFrame.from_dict(ds_f[img_indexes], dtype=str)[num_imgs-checkpointing_steps:]
            split.to_csv(f'{ckp_output_path}/PixelProse_split_{(num_imgs // checkpointing_steps)-1}.csv', index=False)

    if len(counter) > 0:
        with open(f'{stats_output_path}/vocabulary-restart-last_split.json', 'w', encoding='utf-8') as f:
            json.dump(counter, f, ensure_ascii=False)

    # save indexes and stats
    stats = {
        'dataset_size': set_size,
        'downloaded_images': num_imgs,
        'downloads_failed': downloads_failed,
        'image_indexes': img_indexes,
        'unavailable_img_indexes_set': list(unavailable_img_indexes_set),
        'restart_vocabulary': restart_vocabulary,
        'restart_vocabulary_indexes': restart_vocabulary_indexes,
        'words_increment_per_prompt': deltas,
    }
    with open(f'{stats_output_path}/stats.json', 'w', encoding='utf-8') as f:
        json.dump(stats, f, ensure_ascii=False)

    # save dataset csv
    split = pandas.DataFrame.from_dict(ds_f[img_indexes], dtype=str)
    split.to_csv(f'{output_path}/PixelProse_split_{set_size//1000}k.csv', index=False)

    # save dataset parquet
    # ds_f.select(img_indexes).save_to_disk(f'{output_path}/PixelProse_split_{set_size//1000}k')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Download PixelProse subset.")
    parser.add_argument("--output_path", help="output path", type=str, default='./dataset/PixelProse_MVS')
    parser.add_argument("--set_size", help="size of the subset to be sampled and downloaded", type=int, default=400000)
    parser.add_argument("--restart", help="restart from scratch", action='store_true')
    parser.add_argument("--resume_from_checkpoint", help="resume from checkpoint", action='store_true')
    parser.add_argument("--resume_from_rand", help="resume from rand", action='store_true')
    parser.add_argument("--checkpointing_steps", help="checkpointing steps", type=int, default=5000)
    parser.add_argument("--last_ckp", help="last checkpoint to use for resume", type=int, default=0)
    parser.add_argument("--rand_last_ckp_path", help="rand last checkpoint path to use for resume", type=str, default=None)
    args = parser.parse_args()
    assert args.set_size % args.checkpointing_steps == 0, 'set_size should be a multiple of checkpointing_steps ' \
                                                          'to save last checkpoint.'
    main(args)
