from datasets import load_dataset
from collections import Counter
from PIL import Image
from io import BytesIO
from tqdm import tqdm
import json
import urllib.request
import pandas
import os
import shutil
import argparse
import random


def main(args):
    # Load the dataset (~16.9M)
    ds = load_dataset("tomg-group-umd/pixelprose")

    # Filter the dataset (~2.4M)
    ds_f = ds['train'].filter(
        lambda x: x['original_width'] >= 512
                  and x['original_height'] >= 512
                  and x['watermark_class_id'] == 1  # images without watermark (and w/o text)
                  and x['watermark_class_score'] >= 0.8
                  and x['aesthetic_score'] > 4.5
                  and (max(x['original_width'], x['original_height']) / min(x['original_width'], x['original_height'])) <= 1.5  # aspect ratio 3:2/2:3 (or 5:3/3:5)
                  and len(x['vlm_caption'].split('\n')) <= 3  # max 2 '\n'
                  and all(ord(c) >= 0 and ord(c) <= 127 for c in x['vlm_caption'])
                  and not any(c in x['vlm_caption'] for c in r'\|"$%&/()=^*+[]#@<>')
                  and not 'slideplayer' in x['url']
                  and len(x['vlm_caption'].split(',')) <= 80  # max 79 ','
    )

    print(f"Total number of images: {ds_f.num_rows}, subset size: {args.set_size}")

    # Download the subset
    checkpointing_steps = args.checkpointing_steps
    resume_from_checkpoint = args.resume_from_checkpoint
    resume_from_mvs = args.resume_from_mvs
    last_ckp = args.last_ckp
    mvs_last_ckp_path = args.mvs_last_ckp_path
    restart = args.restart

    if restart:
        shutil.rmtree('./dataset/PixelProse_RAND', ignore_errors=True)

    output_path = './dataset/PixelProse_RAND'
    imgs_output_path = f'{output_path}/images'
    csv_output_path = f'{output_path}/csv'
    ckp_output_path = f'{output_path}/checkpoints'
    stats_output_path = f'{output_path}/stats'
    os.makedirs(output_path, exist_ok=True)
    os.makedirs(imgs_output_path, exist_ok=True)
    os.makedirs(csv_output_path, exist_ok=True)
    os.makedirs(ckp_output_path, exist_ok=True)
    os.makedirs(stats_output_path, exist_ok=True)

    headers = {
        "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:72.0) Gecko/20100101 Firefox/72.0"
    }

    if resume_from_checkpoint:
        print(f'Resuming from checkpoint {last_ckp}..')
        with open(f'{ckp_output_path}/checkpoint-{last_ckp}.json', 'r') as f:
            ckp = json.load(f)
        num_imgs = int(ckp['num_imgs'])
        set_size = int(ckp['set_size'])
        img_indexes = list(ckp['img_indexes'])
        unavailable_img_indexes_set = set(ckp['unavailable_img_indexes_set'])
        counter = Counter(ckp['counter'])
        deltas = list(ckp['deltas'])
        #restart_vocabulary = int(ckp['restart_vocabulary'])
        #restart_vocabulary_indexes = list(ckp['restart_vocabulary_indexes'])
        downloads_failed = int(ckp['downloads_failed'])
        #words_list = np.array([set(wl) for wl in ckp['words_list']])
    elif resume_from_mvs:
        print(f'Resuming from mvs {mvs_last_ckp_path}..')
        with open(mvs_last_ckp_path, 'r') as f:
            ckp = json.load(f)
        num_imgs = int(ckp['num_imgs'])
        set_size = int(ckp['set_size']) + args.set_size
        img_indexes = list(ckp['img_indexes'])
        unavailable_img_indexes_set = set(ckp['unavailable_img_indexes_set'])
        counter = Counter()
        deltas = []
        # restart_vocabulary = int(ckp['restart_vocabulary'])
        # restart_vocabulary_indexes = list(ckp['restart_vocabulary_indexes'])
        downloads_failed = int(ckp['downloads_failed'])
        # words_list = np.array([set(wl) for wl in ckp['words_list']])
    else:
        num_imgs = 0
        set_size = args.set_size
        img_indexes = []
        unavailable_img_indexes_set = set()
        counter = Counter()
        deltas = []
        #restart_vocabulary = 0
        #restart_vocabulary_indexes = []
        downloads_failed = 0
        #words_list = np.array(
        #    [{w.lower().strip('_-.,;:!?\'\"\“\”') for w in prompt['vlm_caption'].split()} for prompt in tqdm(ds_f)]
        #    #[{w.lower().strip('_-.,;:!?\'\"\“\”') for w in ds_f[i]['vlm_caption'].split()} for i in tqdm(range(800000))]
        #)

    state = {
        'num_imgs': num_imgs,
        'set_size': set_size,
        'img_indexes': img_indexes,
        'unavailable_img_indexes_set': list(unavailable_img_indexes_set),
        'counter': dict(counter),
        'deltas': deltas,
        #'restart_vocabulary': restart_vocabulary,
        #'restart_vocabulary_indexes': restart_vocabulary_indexes,
        'downloads_failed': downloads_failed,
        #'words_list': [list(ws) for ws in words_list],
    }
    with open(f'{ckp_output_path}/checkpoint-0.json', 'w', encoding='utf-8') as f:
        json.dump(state, f, ensure_ascii=False)

    random.seed(12345)
    indexes = random.sample(range(0, ds_f.num_rows), ds_f.num_rows)

    for idx in indexes:
        # for resumed downloads
        if idx in unavailable_img_indexes_set or idx in set(img_indexes):
            continue

        delta = len({w.lower().strip('_-.,;:!?\'\"\“\”') for w in ds_f[idx]['vlm_caption'].split()} - set(counter.keys()))
        print(f'{num_imgs}: {set_size} | {delta}, {idx}')

        try:
            # print(ds_f[index]['url'])
            request_ = urllib.request.Request(ds_f[idx]['url'], None, headers)  # the assembled request
            response = urllib.request.urlopen(request_, timeout=8)  # store the response
            #if response.info().get_content_maintype() != 'image':
            #    print('not image, skipping..')
            #    raise ValueError('Invalid content')
        except:
            unavailable_img_indexes_set.add(idx)
            downloads_failed += 1
            continue

        imgs_curr_dir = f'{imgs_output_path}/{str(num_imgs // checkpointing_steps).zfill(8)}'
        os.makedirs(imgs_curr_dir, exist_ok=True)

        # check if image is valid
        try:
            img = Image.open(BytesIO(response.read()))
            img.save(f"{imgs_curr_dir}/{ds_f[idx]['key']}.jpg")
        except:
            unavailable_img_indexes_set.add(idx)
            downloads_failed += 1
            print('failed image open, skipping..')
            continue

        img_indexes.append(idx)
        counter.update(Counter([w.lower().strip('_-.,;:!?\'\"\“\”') for w in ds_f[idx]['vlm_caption'].split()]))
        deltas.append(delta)
        num_imgs += 1

        if (num_imgs % checkpointing_steps) == 0:
            state = {
                'num_imgs': num_imgs,
                'set_size': set_size,
                'img_indexes': img_indexes,
                'unavailable_img_indexes_set': list(unavailable_img_indexes_set),
                'counter': dict(counter),
                'deltas': deltas,
                #'restart_vocabulary': restart_vocabulary,
                #'restart_vocabulary_indexes': restart_vocabulary_indexes,
                'downloads_failed': downloads_failed,
                #'words_list': [list(ws) for ws in words_list],
            }
            with open(f'{ckp_output_path}/checkpoint-{num_imgs // checkpointing_steps}.json', 'w', encoding='utf-8') as f:
                json.dump(state, f, ensure_ascii=False)
            # save dataset split csv
            split = pandas.DataFrame.from_dict(ds_f[img_indexes], dtype=str)[num_imgs-checkpointing_steps:]
            split.to_csv(f'{csv_output_path}/PixelProse_split_{(num_imgs // checkpointing_steps)-1}.csv', index=False)

        if num_imgs == set_size:
            break

    with open(f'{stats_output_path}/vocabulary_split.json', 'w', encoding='utf-8') as f:
        json.dump(counter, f, ensure_ascii=False)

    # save indexes and stats
    stats = {
        'dataset_size': set_size,
        'downloaded_images': num_imgs,
        'downloads_failed': downloads_failed,
        'image_indexes': img_indexes,
        'unavailable_img_indexes_set': list(unavailable_img_indexes_set),
        #'restart_vocabulary': restart_vocabulary,
        #'restart_vocabulary_indexes': restart_vocabulary_indexes,
        'words_increment_per_prompt': deltas,
    }
    with open(f'{stats_output_path}/stats.json', 'w', encoding='utf-8') as f:
        json.dump(stats, f, ensure_ascii=False)

    # save dataset csv
    split = pandas.DataFrame.from_dict(ds_f[img_indexes], dtype=str)
    split.to_csv(f'{output_path}/PixelProse_split_{set_size//1000}k.csv', index=False)

    # save dataset parquet
    # ds_f.select(img_indexes).save_to_disk(f'{output_path}/PixelProse_split_{set_size//1000}k')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Download PixelProse subset.")
    parser.add_argument("--output_path", help="output path", type=str, default='./dataset/PixelProse_RAND')
    parser.add_argument("--set_size", help="size of the subset to be sampled and downloaded", type=int, default=200000)
    parser.add_argument("--restart", help="restart from scratch", action='store_true')
    parser.add_argument("--resume_from_checkpoint", help="resume from checkpoint", action='store_true')
    parser.add_argument("--resume_from_mvs", help="resume from mvs", action='store_true')
    parser.add_argument("--checkpointing_steps", help="checkpointing steps", type=int, default=5000)
    parser.add_argument("--last_ckp", help="last checkpoint to use for resume", type=int, default=0)
    parser.add_argument("--mvs_last_ckp_path", help="mvs last checkpoint path to use for resume", type=str, default=None)
    args = parser.parse_args()
    assert args.set_size % args.checkpointing_steps == 0, 'set_size should be a multiple of checkpointing_steps ' \
                                                          'to save last checkpoint.'
    main(args)
