import os
import torch
import argparse
import json
from diffusers import AutoencoderKL, StableDiffusionPipeline, DPMSolverMultistepScheduler
from sg_diffuser.types import SceneGraph
from sg_diffuser.sg_diffuser import SceneGraphEncoder, SGDiffuser


def main(args):
    if args.use_text:
        prompt = args.prompt
    else:
        with open(args.prompt, 'r') as f:
            data = json.load(f)

        nodes = {**data["EL_dict"], **data["rel_dict"]}
        sorted_nodes = sorted(nodes, key=nodes.get)
        edge_index = torch.tensor(data["adj_list"]).t().contiguous()

        prompt = SceneGraph(
            nodes=sorted_nodes,
            adj_list=edge_index,
            node_count=len(sorted_nodes),
            edge_count=edge_index.shape[1]
        )

    vae = AutoencoderKL.from_pretrained('stabilityai/sd-vae-ft-mse', torch_dtype=torch.float16)

    pipeline = StableDiffusionPipeline.from_pretrained(
        'SG161222/Realistic_Vision_V5.1_noVAE',
        vae=vae,
        text_encoder=None,
        tokenizer=None,
        # unet=unet,
        safety_checker=None,
        torch_dtype=torch.float16,
    )

    pipeline.scheduler = DPMSolverMultistepScheduler.from_config(
        pipeline.scheduler.config,
        algorithm_type="sde-dpmsolver++",
        use_karras_sigmas='true'
    )

    sg_model = SGDiffuser(
        sd_pipe=pipeline,
        sg_encoder=SceneGraphEncoder(),
        text_encoder_path=None,
        tokenizer=None,
        proj_type='GCN',
        ip_ckpt='sg-diffuser-model-ckpt-1k-lr1e-4/checkpoint-40000/model.safetensors',
        use_text=args.use_text,
        device=args.device
    )

    image = sg_model.generate(
        prompt=prompt,
        #prompt="This image displays: A young woman and a young man are standing in a kitchen. The woman has bright blonde hair that is cut in a choppy style, with bangs that are swept to the side. She is wearing a blue t-shirt. The man is standing next to her, with his arm around her shoulder. He has short dark brown hair and is wearing a red sweatshirt. He has a small smile on his face, and they appear to be posing for the picture. The background of the image is blurry, but it looks like there is a breakfast bar behind them."
        num_inference_steps=30,
        guidance_scale=0,
        height=512,
        width=512,
        seed=args.seed
    )[0]

    image.save('test.jpg')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Scene Graph Diffuser.")
    parser.add_argument("--prompt", help="prompt (string or path to scene graph)", type=str, default='./dataset/val_prompts/001144312.json')
    parser.add_argument("--use_text", help="use text prompt", action='store_true')
    parser.add_argument("--seed", help="seed", type=int, default=12345)
    parser.add_argument("--device", help="device", type=str, default='cuda:0')
    args = parser.parse_args()
    main(args)
