import torch
import json
from transformers import AutoTokenizer, AutoModelForCausalLM
from transformers.utils.import_utils import _is_package_available
from .sg_processor import SceneGraphProcessor


class PromptParser:
    def __init__(
            self,
            ckpt='meta-llama/Meta-Llama-3.1-8B-Instruct',
            use_instruct_prompt=True,
            torch_compile=False,
            device='cuda:0'
    ):
        self.model = AutoModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path=ckpt,
            torch_dtype=torch.bfloat16,
            attn_implementation="flash_attention_2" if _is_package_available("flash-attn") else "sdpa",
            device_map=device,
        )
        self.tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=ckpt,
            padding_side='left'
        )
        self.tokenizer.pad_token_id = self.tokenizer.eos_token_id

        if torch_compile:
            torch.backends.cuda.matmul.allow_tf32 = True
            self.model.generation_config.max_length = 3*3072
            self.model.generation_config.cache_implementation = "static"
            self.model.forward = torch.compile(self.model.forward, mode="reduce-overhead", fullgraph=True)

        self.terminators = [
            self.tokenizer.eos_token_id,
            self.tokenizer.convert_tokens_to_ids("<|eot_id|>")
        ]

        self.use_instruct_prompt = use_instruct_prompt
        self.prompt_file = (
            './parser/prompts/prompt_instruct.json' if use_instruct_prompt else './parser/prompts/prompt.json'
        )
        with open(self.prompt_file, 'r') as f:
            self.messages = json.load(f)

        self.easier_prompt_file = (
            './parser/prompts/prompt_instruct_easier.json' if use_instruct_prompt else './parser/prompts/prompt_easier.json'
        )
        with open(self.easier_prompt_file, 'r') as f:
            self.easier_messages = json.load(f)

        self.post_processor = SceneGraphProcessor(self)

    def __call__(
            self,
            caption,
            temp=0.4,
            postprocess=False,
            use_easier_prompt=False
    ):
        prompt_messages = self.messages if not use_easier_prompt else self.easier_messages

        if self.use_instruct_prompt:
            if isinstance(caption, str):
                caption = caption.strip('"').strip()
                messages = [prompt_messages + [{"role": "user", "content": f'Caption: "{caption}"'}]]
            else:
                caption = [cap.strip('"').strip() for cap in caption]
                messages = [prompt_messages + [{"role": "user", "content": f'Caption: "{cap}"'}] for cap in caption]
        else:
            if isinstance(caption, str):
                caption = caption.strip('"').strip()
                messages = [[{"role": "system", "content": prompt_messages["content"] + f'"{caption}"'}]]
            else:
                caption = [cap.strip('"').strip() for cap in caption]
                messages = [[{"role": "system", "content": prompt_messages["content"] + f'"{cap}"'}] for cap in caption]

        texts = self.tokenizer.apply_chat_template(
            messages,
            add_generation_prompt=True,
            tokenize=False,
        )

        #
        inputs = self.tokenizer(texts, padding='longest', return_tensors="pt")
        inputs = {key: val.to(self.model.device) for key, val in inputs.items()}
        temp_texts = self.tokenizer.batch_decode(inputs["input_ids"], skip_special_tokens=True)
        #

        #with torch.backends.cuda.sdp_kernel(enable_flash=True, enable_math=False, enable_mem_efficient=False):
        outputs = self.model.generate(
            **inputs,
            #inputs['input_ids'],
            eos_token_id=self.terminators,
            do_sample=True,
            temperature=temp,
            top_p=0.9,
            pad_token_id=self.tokenizer.eos_token_id,
            #attention_mask=inputs["attention_mask"],
            max_new_tokens=1000
        )

        gen_text = self.tokenizer.batch_decode(outputs, skip_special_tokens=True)
        gen_text = [i[len(temp_texts[idx]):] for idx, i in enumerate(gen_text)]

        if postprocess:
            # check if bs=1, call SceneGraphProcessor and return result (SceneGraph obj)
            assert len(messages) == 1, "Expected batch size == 1"
            scene_graph = self.post_processor(*gen_text, caption)
            return scene_graph
        else:
            return gen_text
