import torch
import json
from sg_diffuser.types import SceneGraph


class SceneGraphProcessor:

    def __init__(self, llm_parser):
        self.llm_parser = llm_parser

    def __format_list_in_triplets(self, input_list):
        first_element = input_list[0]
        middle_elements = " ".join(input_list[1:-1])
        last_element = input_list[-1]
        output_list = [first_element, middle_elements, last_element]
        return output_list

    def _post_process(self, gen_text, try_recover):
        res = None
        try:
            # try to read the LLM raw output
            res = json.loads('{' + gen_text + '}')
            if max(len(triplet) for triplet in res['Knowledge Graph']) > 3:
                raise ValueError('Triplets with more than 3 elements.')
        except Exception as e:
            if not try_recover:
                print(f'Error in the LLM output format ({e}), trying again..')
                return None
            elif res is None:
                # error in the format nothing to correct (failed)
                print('Error in the LLM output format, failed to recover. Skipped.')
                return None
            else:
                print('Correction required due to triplets with more than 3 elements. Applying corrections..')

        # apply corrections if any (i.e., join elements in triplets with more than 3 elements) and create the final dict
        KG = res['Knowledge Graph']
        index = 0
        EL_list = set()
        EL_dict = {}
        rel_dict = {}
        adj_list = []

        # Triplet Formatting, if len(triplet)>3 ---> len (triplet)==3 (if any)
        for i in range(0, len(KG)):
            if len(KG[i]) > 3:
                KG[i] = self.__format_list_in_triplets(KG[i])

        # Entity List extraction
        for triplet in KG:
            EL_list.add(triplet[0])
            EL_list.add(triplet[-1])
        EL_list = list(EL_list) if len(EL_list) else res['Entity List']

        # Entity Mapping - Triplets with two elements have both elemets mapped in the EL_dict
        for triplet in KG:
            if triplet[0] not in EL_dict:
                EL_dict[triplet[0]] = index
                index += 1
            if triplet[-1] not in EL_dict:
                EL_dict[triplet[-1]] = index
                index += 1

        # handle graphs with unconnected nodes (empty KG - no relationships)
        if len(KG) == 0 and len(EL_list) > 0:
            for entity in EL_list:
                EL_dict[entity] = index
                index += 1

        # Relationship mapping if len == 3
        for triplet in KG:
            if len(triplet) == 3:
                base_relation = triplet[1]
                found = False
                max_ind = -1
                for key in rel_dict.keys():
                    if key.startswith(base_relation + "_"):
                        found = True
                        current_index = int(key.split("_")[-1])
                        if current_index > max_ind:
                            max_ind = current_index
                if not found:
                    triplet[1] = base_relation + "_0"
                    rel_dict[triplet[1]] = index
                else:
                    new_index = max_ind + 1
                    triplet[1] = base_relation + f"_{new_index}"
                    rel_dict[triplet[1]] = index
                index += 1

        # Adjacency List creation
        for triplet in KG:
            if len(triplet) == 3:
                adj_list.append((EL_dict[triplet[0]], rel_dict[triplet[1]]))
                adj_list.append((rel_dict[triplet[1]], EL_dict[triplet[2]]))
            elif len(triplet) == 2:
                adj_list.append((EL_dict[triplet[0]], EL_dict[triplet[1]]))

        return {
            'EL_list': EL_list,
            'rel_list': list(rel_dict.keys()),
            "KG": KG,
            "EL_dict": EL_dict,
            "rel_dict": rel_dict,
            "adj_list": adj_list,
        }

    def __call__(self, gen_text, caption, max_attempts=3, save_path=None, filename=None):
        # try postprocessing
        result = self._post_process(gen_text, try_recover=False)

        # if failed, try again postprocessing for max_attempts times
        if result is None:
            # by default tries 2 more times using the same prompt before using a downgraded version of the prompt
            for i in range(max_attempts):
                is_last_attempt = (i == (max_attempts - 1))
                gen_text = self.llm_parser(caption, use_easier_prompt=is_last_attempt)[0]
                result = self._post_process(gen_text, try_recover=is_last_attempt)
                if result is not None:
                    break

        # save results in a dict (if correctly processed)
        if save_path is not None and result is not None:
            with open(f'{save_path}/{filename}.json', 'w') as f:
                json.dump(result, f)

        # return None if postprocess failed
        if result is None:
            return None

        # build a Scene Graph object to return
        nodes = {**result['rel_dict'], **result['EL_dict']}
        sorted_nodes = sorted(nodes, key=nodes.get)

        adj_list = result['adj_list']
        edge_index = (
            torch.tensor(adj_list).t().contiguous() if len(adj_list) > 0 else torch.tensor([[], []], dtype=torch.int64)
        )

        return SceneGraph(
                nodes=sorted_nodes,
                adj_list=edge_index,
                node_count=len(sorted_nodes),
                edge_count=edge_index.shape[1]
            )
