import os
from typing import List
from pathlib import Path
import torch
from safetensors import safe_open
from transformers import CLIPTokenizer, AutoTokenizer, CLIPTextModelWithProjection
from .utils import is_torch2_available, get_generator

if is_torch2_available():
    from diffusers.models.attention_processor import (
        CustomDiffusionAttnProcessor2_0 as CustomDiffusionAttnProcessor,
        AttnProcessor2_0 as AttnProcessor,
    )
else:
    from diffusers.models.attention_processor import CustomDiffusionAttnProcessor, AttnProcessor

from parser.llm_parser import PromptParser
from .adapters import GCNAdapter, GATAdapter, Resampler
from .types import SceneGraph


class SceneGraphEncoder:
    def __init__(
            self,
            text_encoder_path='openai/clip-vit-large-patch14',
            tokenizer=None,
            device='cuda:0'
    ):
        self.device = device
        self.text_encoder_path = text_encoder_path

        if isinstance(text_encoder_path, (str, Path)):
            self.text_encoder = CLIPTextModelWithProjection.from_pretrained(self.text_encoder_path).to(
                self.device, dtype=torch.float16
            )
            self.tokenizer = CLIPTokenizer.from_pretrained(self.text_encoder_path, clean_up_tokenization_spaces=True)
        else:
            self.text_encoder = text_encoder_path
            self.tokenizer = tokenizer

    @torch.no_grad()
    def __call__(
            self,
            scene_graph,
            padding,
            do_classifier_free_guidance
    ):
        if isinstance(scene_graph, SceneGraph):
            batch_size = 1
            scene_graph = [scene_graph]
        elif isinstance(scene_graph, List) and all(isinstance(sg, SceneGraph) for sg in scene_graph):
            batch_size = len(scene_graph)
        else:
            raise ValueError("Expected SceneGraph or List[SceneGraph]!")

        if padding:
            max_n_nodes = max(len(sg.nodes) for sg in scene_graph)

        scene_graph_embeds = []
        uncond_scene_graph_embeds = []
        for sg in scene_graph:
            nodes = sg.nodes
            if padding:
                nodes.extend([""] * (max_n_nodes-len(nodes)))

            inputs_ids = self.tokenizer(nodes, padding=True, return_tensors="pt").input_ids
            outputs = self.text_encoder(inputs_ids.to(self.device))
            scene_graph_embeds.append(outputs.text_embeds)

            if do_classifier_free_guidance:
                empty_nodes = [""] * len(nodes)
                uncond_inputs_ids = self.tokenizer(empty_nodes, padding=True, return_tensors="pt").input_ids
                uncond_outputs = self.text_encoder(uncond_inputs_ids.to(self.device))
                uncond_scene_graph_embeds.append(uncond_outputs.text_embeds)

        if padding:
            scene_graph_embeds = torch.stack(scene_graph_embeds, dim=0)
            uncond_scene_graph_embeds = (
                torch.stack(uncond_scene_graph_embeds, dim=0) if len(uncond_scene_graph_embeds) > 0 else None
            )
        else:
            scene_graph_embeds = torch.cat(scene_graph_embeds, dim=0)
            uncond_scene_graph_embeds = (
                torch.cat(uncond_scene_graph_embeds, dim=0) if len(uncond_scene_graph_embeds) > 0 else None
            )

        return scene_graph_embeds, uncond_scene_graph_embeds

    @torch.no_grad()
    def get_empty_node_embed(self):
        ids = self.tokenizer([''], padding=True, return_tensors="pt").input_ids
        embed = self.text_encoder(ids.to(self.device)).text_embeds
        return embed


class SGDiffuser:
    def __init__(
            self,
            sd_pipe,
            sg_encoder,
            text_encoder_path,
            tokenizer,
            proj_type,
            use_text,
            ip_ckpt,
            device,
    ):
        self.text_encoder_path = text_encoder_path
        self.proj_type = proj_type
        self.use_text = use_text
        self.ip_ckpt = ip_ckpt
        self.device = device

        # load pipe
        self.pipe = sd_pipe.to(self.device, dtype=torch.float16)
        #self.set_cross_attn()

        # load prompt parser
        self.prompt_parser = PromptParser(device=self.device) if use_text else None
        # load scene graph encoder
        if isinstance(sg_encoder, SceneGraphEncoder):
            self.scene_graph_encoder = sg_encoder
        else:
            self.scene_graph_encoder = SceneGraphEncoder(
                text_encoder_path=self.text_encoder_path, tokenizer=tokenizer, device=self.device
            )
        # load scene graph proj model
        if isinstance(proj_type, str):
            self.sg_proj_model = self.init_proj(self.proj_type)
        else:
            self.sg_proj_model = proj_type

        #self.load_adapter()
        if ip_ckpt is not None:
            self.set_cross_attn()
            self.load_adapter()

    def init_proj(self, proj_type):
        if proj_type == 'GCN':
            proj_model = GCNAdapter(
                input_channels=self.pipe.unet.config.cross_attention_dim,
                hidden_channels=self.pipe.unet.config.cross_attention_dim,
                output_channels=self.pipe.unet.config.cross_attention_dim,
            ).to(self.device, dtype=torch.float16)
        elif proj_type == 'GAT':
            proj_model = GATAdapter()
        elif proj_type == 'Resampler':
            proj_model = Resampler(
                dim=self.pipe.unet.config.cross_attention_dim,
                depth=4,
                dim_head=64,
                heads=12,
                num_queries=77,
                embedding_dim=self.pipe.unet.config.cross_attention_dim,
                output_dim=self.pipe.unet.config.cross_attention_dim,
                ff_mult=4,
            ).to(self.device, dtype=torch.float16)
        else:
            raise ValueError("Adapter model not supported!")

        return proj_model

    def set_cross_attn(self, train_kv=True, train_q_out=True):
        unet = self.pipe.unet
        attn_procs = {}
        for name in unet.attn_processors.keys():
            cross_attention_dim = None if name.endswith("attn1.processor") else unet.config.cross_attention_dim
            if name.startswith("mid_block"):
                hidden_size = unet.config.block_out_channels[-1]
            elif name.startswith("up_blocks"):
                block_id = int(name[len("up_blocks.")])
                hidden_size = list(reversed(unet.config.block_out_channels))[block_id]
            elif name.startswith("down_blocks"):
                block_id = int(name[len("down_blocks.")])
                hidden_size = unet.config.block_out_channels[block_id]
            if cross_attention_dim is None:
                #attn_procs[name] = AttnProcessor()
                attn_procs[name] = CustomDiffusionAttnProcessor(
                    train_kv=False,
                    train_q_out=False,
                    hidden_size=hidden_size,
                    cross_attention_dim=cross_attention_dim,
                )
            else:
                attn_procs[name] = CustomDiffusionAttnProcessor(
                    train_kv=train_kv,
                    train_q_out=train_q_out,
                    hidden_size=hidden_size,
                    cross_attention_dim=cross_attention_dim,
                ).to(self.device, dtype=torch.float16)

        unet.set_attn_processor(attn_procs)

    def load_adapter(self):
        if os.path.splitext(self.ip_ckpt)[-1] == ".safetensors":
            state_dict = {"sg_proj_model": {}, "adapter_modules": {}, "unet": {}}
            with safe_open(self.ip_ckpt, framework="pt", device="cpu") as f:
                for key in f.keys():
                    if key.startswith("sg_proj_model."):
                        state_dict["sg_proj_model"][key.replace("sg_proj_model.", "")] = f.get_tensor(key)
                    elif key.startswith("adapter_modules."):
                        state_dict["adapter_modules"][key.replace("adapter_modules.", "")] = f.get_tensor(key)
                    #elif "custom_diffusion" in key:
                        # fix for removed shared tensors :{
                    #    state_dict["adapter_modules"][key.replace("unet.", "")] = f.get_tensor(key)
                    elif key.startswith("unet."):
                        state_dict["unet"][key.replace("unet.", "")] = f.get_tensor(key)
        else:
            state_dict = torch.load(self.ip_ckpt, map_location="cpu")
        self.sg_proj_model.load_state_dict(state_dict["sg_proj_model"])
        self.pipe.unet.load_state_dict(state_dict["unet"])
        #ip_layers = torch.nn.ModuleList(self.pipe.unet.attn_processors.values())
        #ip_layers.load_state_dict(state_dict["adapter_modules"])

    def generate(
        self,
        prompt,
        negative_prompt=None,
        #num_samples=4,
        seed=None,
        guidance_scale=7.5,
        num_inference_steps=30,
        **kwargs,
    ):

        if self.use_text and not isinstance(prompt, str):
            raise ValueError("Expected string as prompt!")
        elif not self.use_text and not isinstance(prompt, SceneGraph):
            raise ValueError("Expected SceneGraph object as prompt!")

        scene_graph = self.prompt_parser(prompt, postprocess=True) if self.use_text else prompt

        scene_graph_embeds, uncond_scene_graph_embeds = self.scene_graph_encoder(
            scene_graph=scene_graph,
            padding=isinstance(self.sg_proj_model, Resampler),
            do_classifier_free_guidance=guidance_scale > 1.0 and negative_prompt is None,
        )

        if guidance_scale > 1.0 and negative_prompt is not None:
            negative_scene_graph = self.prompt_parser(negative_prompt) if self.use_text else negative_prompt

            uncond_scene_graph_embeds, _ = self.scene_graph_encoder(
                scene_graph=negative_scene_graph,
                padding=isinstance(self.sg_proj_model, Resampler),
                do_classifier_free_guidance=False,
            )

        scene_graph_embeds = self.sg_proj_model(
            scene_graph_embeds, scene_graph.adj_list.to(self.device)
        )
        uncond_scene_graph_embeds = self.sg_proj_model(
            uncond_scene_graph_embeds, torch.tensor([[], []], dtype=torch.int64).to(self.device)
        ) if uncond_scene_graph_embeds is not None else None

        if len(scene_graph_embeds.shape) == 2:
            scene_graph_embeds = scene_graph_embeds.unsqueeze(0)
        if uncond_scene_graph_embeds is not None and len(uncond_scene_graph_embeds.shape) == 2:
            uncond_scene_graph_embeds = uncond_scene_graph_embeds.unsqueeze(0)
        generator = get_generator(seed, self.device)

        images = self.pipe(
            prompt_embeds=scene_graph_embeds,
            negative_prompt_embeds=uncond_scene_graph_embeds,
            guidance_scale=guidance_scale,
            num_inference_steps=num_inference_steps,
            generator=generator,
            **kwargs,
        ).images

        return images
