import os
#import random
import argparse
from pathlib import Path
import json
import itertools
import numpy as np
#import time
from datetime import datetime
from tqdm import tqdm
import math
import shutil
import torch
import torch.nn.functional as F
#from torchvision import transforms
from PIL import Image
#from einops import rearrange
#from transformers import CLIPImageProcessor
from accelerate import Accelerator
from accelerate.logging import get_logger
from accelerate.utils import ProjectConfiguration, set_seed
from diffusers import (
    AutoencoderKL, DDPMScheduler, UNet2DConditionModel, StableDiffusionPipeline, DPMSolverMultistepScheduler
)
from diffusers.utils import is_wandb_available
from diffusers.optimization import get_scheduler

from sg_diffuser.adapters import Resampler, GCNAdapter, GATAdapter
from sg_diffuser.utils import is_torch2_available
from sg_diffuser.types import SceneGraph
from sg_diffuser.sg_diffuser import SceneGraphEncoder, SGDiffuser
if is_torch2_available():
    from diffusers.models.attention_processor import (
        CustomDiffusionAttnProcessor2_0 as CustomDiffusionAttnProcessor,
        AttnProcessor2_0 as AttnProcessor
    )
else:
    from diffusers.models.attention_processor import CustomDiffusionAttnProcessor, AttnProcessor

from dataset.dataset import CustomDataset, collate_fn, get_transforms_rand_crop

if is_wandb_available():
    import wandb


logger = get_logger(__name__)


class TrainAdapterHelper(torch.nn.Module):
    """TrainAdapterHelper"""
    def __init__(
            self,
            unet,
            sg_proj_model,
            adapter_modules,
            ckpt_path=None
    ):
        super().__init__()
        self.unet = unet
        self.sg_proj_model = sg_proj_model
        self.adapter_modules = adapter_modules

        if ckpt_path is not None:
            self.load_from_checkpoint(ckpt_path)

    def forward(
            self,
            noisy_latents,
            timesteps,
            scene_graph_embeds,
            edge_index,
            node_counts,
            empty_node_embed
    ):
        # Pad tokens before for transformer projection
        if isinstance(self.sg_proj_model, Resampler) and scene_graph_embeds.shape[0] != len(node_counts):
            scene_graph_embeds = scene_graph_embeds.squeeze(0)
            scene_graph_embeds = [torch.cat([scene_graph_embeds[sum(node_counts[:i]):sum(node_counts[:i]) + n, :],
                                             empty_node_embed.repeat(max(node_counts) - n, 1)], dim=0)
                                  for i, n in enumerate(node_counts)]
            scene_graph_embeds = torch.stack(scene_graph_embeds, dim=0)

        encoder_hidden_states = self.sg_proj_model(scene_graph_embeds, edge_index)

        # Pad tokens after for gnn projection
        if len(encoder_hidden_states.shape) == 2:
            hidden_states = [torch.cat([encoder_hidden_states[sum(node_counts[:i]):sum(node_counts[:i]) + n, :],
                                        empty_node_embed.repeat(max(node_counts) - n, 1)], dim=0)
                             for i, n in enumerate(node_counts)]
            encoder_hidden_states = torch.stack(hidden_states, dim=0)

        # Predict the noise residual
        noise_pred = self.unet(noisy_latents, timesteps, encoder_hidden_states).sample
        return noise_pred

    def load_from_checkpoint(self, ckpt_path: str):
        # Calculate original checksums
        orig_sg_proj_sum = torch.sum(torch.stack([torch.sum(p) for p in self.sg_proj_model.parameters()]))
        orig_adapter_sum = torch.sum(torch.stack([torch.sum(p) for p in self.adapter_modules.parameters()]))

        state_dict = torch.load(ckpt_path, map_location="cpu")

        # Load state dict for image_proj_model and adapter_modules
        self.sg_proj_model.load_state_dict(state_dict["sg_proj_model"], strict=True)
        self.adapter_modules.load_state_dict(state_dict["adapter_modules"], strict=True)

        # Calculate new checksums
        new_sg_proj_sum = torch.sum(torch.stack([torch.sum(p) for p in self.sg_proj_model.parameters()]))
        new_adapter_sum = torch.sum(torch.stack([torch.sum(p) for p in self.adapter_modules.parameters()]))

        # Verify if the weights have changed
        assert orig_sg_proj_sum != new_sg_proj_sum, "Weights of sg_proj_model did not change!"
        assert orig_adapter_sum != new_adapter_sum, "Weights of adapter_modules did not change!"

        print(f"Successfully loaded weights from checkpoint {ckpt_path}")


def parse_args():
    parser = argparse.ArgumentParser(description="Simple example of a training script.")
    parser.add_argument(
        "--pretrained_model_name_or_path",
        type=str,
        default=None,
        required=True,
        help="Path to pretrained model or model identifier from huggingface.co/models.",
    )
    parser.add_argument(
        "--pretrained_vae_model_name_or_path",
        type=str,
        default=None,
        help=(
            "Path to an improved VAE to stabilize training. "
            "For more details check out: https://github.com/huggingface/diffusers/pull/4038."
        ),
    )
    parser.add_argument(
        "--pretrained_sg_adapter_path",
        type=str,
        default=None,
        help="Path to pretrained sg adapter model. If not specified weights are initialized randomly.",
    )
    parser.add_argument(
        "--proj_type",
        type=str,
        default="GCN",
        help="The adapter model to be used.",
    )
    parser.add_argument(
        "--revision",
        type=str,
        default=None,
        required=False,
        help=(
            "Revision of pretrained model identifier from huggingface.co/models. Trainable model components should be"
            " float32 precision."
        ),
    )
    parser.add_argument(
        "--num_tokens",
        type=int,
        default=77,
        help="Number of tokens to query from the Scene Graph node encoding (used by Resampler).",
    )
    parser.add_argument(
        "--data_root_path",
        type=str,
        default=None,
        required=True,
        help="Training data annotations (json) root path",
    )
    parser.add_argument(
        "--images_root_path",
        type=str,
        default=None,
        required=True,
        help="Training images (jpg) root path",
    )
    parser.add_argument(
        "--text_encoder_path",
        type=str,
        default='openai/clip-vit-large-patch14',
        help="Path to CLIP text encoder",
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        default="sg-diffuser-model-ckpt",
        help="The output directory where the model predictions and checkpoints will be written.",
    )
    parser.add_argument(
        "--logging_dir",
        type=str,
        default="logs",
        help=(
            "[TensorBoard](https://www.tensorflow.org/tensorboard) log directory. Will default to"
            " *output_dir/runs/**CURRENT_DATETIME_HOSTNAME***."
        ),
    )
    parser.add_argument("--seed", type=int, default=None, help="A seed for reproducible training.")
    parser.add_argument(
        "--resolution",
        type=int,
        default=512,
        help="The resolution for input images.",
    )
    parser.add_argument(
        "--gradient_accumulation_steps",
        type=int,
        default=1,
        help="Number of updates steps to accumulate before performing a backward/update pass.",
    )
    parser.add_argument(
        "--gradient_checkpointing",
        action="store_true",
        help="Whether or not to use gradient checkpointing to save memory at the expense of slower backward pass.",
    )
    parser.add_argument(
        "--learning_rate",
        type=float,
        default=1e-4,
        help="Learning rate to use.",
    )
    parser.add_argument("--adam_beta1", type=float, default=0.9, help="The beta1 parameter for the Adam optimizer.")
    parser.add_argument("--adam_beta2", type=float, default=0.999, help="The beta2 parameter for the Adam optimizer.")
    parser.add_argument("--weight_decay", type=float, default=1e-2, help="Weight decay to use.")
    parser.add_argument("--adam_epsilon", type=float, default=1e-08, help="Epsilon value for the Adam optimizer")
    parser.add_argument("--max_grad_norm", default=1.0, type=float, help="Max gradient norm.")
    parser.add_argument(
        "--scale_lr",
        action="store_true",
        default=False,
        help="Scale the learning rate by the number of GPUs, gradient accumulation steps, and batch size.",
    )
    parser.add_argument(
        "--lr_scheduler",
        type=str,
        default="constant",
        help=(
            'The scheduler type to use. Choose between ["linear", "cosine", "cosine_with_restarts", "polynomial",'
            ' "constant", "constant_with_warmup"]'
        ),
    )
    parser.add_argument(
        "--lr_warmup_steps",
        type=int,
        default=500,
        help="Number of steps for the warmup in the lr scheduler."
    )
    parser.add_argument(
        "--lr_num_cycles",
        type=int,
        default=1,
        help="Number of hard resets of the lr in cosine_with_restarts scheduler.",
    )
    parser.add_argument("--lr_power", type=float, default=1.0, help="Power factor of the polynomial scheduler.")
    parser.add_argument("--num_train_epochs", type=int, default=100)
    parser.add_argument(
        "--max_train_steps",
        type=int,
        default=None,
        help="Total number of training steps to perform.  If provided, overrides num_train_epochs.",
    )
    parser.add_argument(
        "--train_batch_size",
        type=int,
        default=8,
        help="Batch size (per device) for the training dataloader."
    )
    parser.add_argument(
        "--dataloader_num_workers",
        type=int,
        default=0,
        help=(
            "Number of subprocesses to use for data loading. 0 means that the data will be loaded in the main process."
        ),
    )
    '''parser.add_argument(
        "--save_steps",
        type=int,
        default=2000,
        help=(
            "Save a checkpoint of the training state every X updates"
        ),
    )'''
    parser.add_argument(
        "--checkpointing_steps",
        type=int,
        default=2000,
        help=(
            "Save a checkpoint of the training state every X updates. Checkpoints can be used for resuming training via `--resume_from_checkpoint`. "
            "In the case that the checkpoint is better than the final trained model, the checkpoint can also be used for inference."
            "Using a checkpoint for inference requires separate loading of the original pipeline and the individual checkpointed model components."
            "See https://huggingface.co/docs/diffusers/main/en/training/dreambooth#performing-inference-using-a-saved-checkpoint for step by step"
            "instructions."
        ),
    )
    parser.add_argument(
        "--checkpoints_total_limit",
        type=int,
        default=None,
        help="Max number of checkpoints to store.",
    )
    parser.add_argument(
        "--resume_from_checkpoint",
        type=str,
        default=None,
        help=(
            "Whether training should be resumed from a previous checkpoint. Use a path saved by"
            ' `--checkpointing_steps`, or `"latest"` to automatically select the last available checkpoint.'
        ),
    )
    parser.add_argument(
        "--mixed_precision",
        type=str,
        default=None,
        choices=["no", "fp16", "bf16"],
        help=(
            "Whether to use mixed precision. Choose between fp16 and bf16 (bfloat16). Bf16 requires PyTorch >="
            " 1.10.and an Nvidia Ampere GPU.  Default to the value of accelerate config of the current system or the"
            " flag passed with the `accelerate.launch` command. Use this argument to override the accelerate config."
        ),
    )
    parser.add_argument(
        "--set_grads_to_none",
        action="store_true",
        help=(
            "Save more memory by using setting grads to None instead of zero. Be aware, that this changes certain"
            " behaviors, so disable this argument if it causes any problems. More info:"
            " https://pytorch.org/docs/stable/generated/torch.optim.Optimizer.zero_grad.html"
        ),
    )
    parser.add_argument(
        "--report_to",
        type=str,
        default="tensorboard",
        help=(
            'The integration to report the results and logs to. Supported platforms are `"tensorboard"`'
            ' (default), `"wandb"` and `"comet_ml"`. Use `"all"` to report to all integrations.'
        ),
    )
    parser.add_argument(
        "--tracker_log_dir",
        type=str,
        default="./tracker_logs",
        help=(
            "The `log_dir` argument passed to Accelerator.init_trackers for"
            " more information see https://huggingface.co/docs/accelerate/v0.17.0/en/package_reference/accelerator#accelerate.Accelerator"
        ),
    )
    parser.add_argument(
        "--tracker_project_name",
        type=str,
        default="SceneGraph-Diffuser",
        help=(
            "The `project_name` argument passed to Accelerator.init_trackers for"
            " more information see https://huggingface.co/docs/accelerate/v0.17.0/en/package_reference/accelerator#accelerate.Accelerator"
        ),
    )
    parser.add_argument(
        "--tracker_run_name",
        type=str,
        default="run-{TIME}",
        help=(
            "The `run_name` argument passed to Accelerator.init_trackers for"
            " more information see https://huggingface.co/docs/accelerate/v0.17.0/en/package_reference/accelerator#accelerate.Accelerator"
        ),
    )
    parser.add_argument(
        "--proportion_empty_prompts",
        type=float,
        default=0,
        help="Proportion of image prompts to be replaced with empty strings. Defaults to 0 (no prompt replacement).",
    )
    parser.add_argument(
        "--validation_prompt",
        type=str,
        default=None,
        nargs="+",
        help="A set of prompts evaluated every `--validation_steps` and logged to `--report_to`.",
    )
    parser.add_argument(
        "--validation_image",
        type=str,
        default=None,
        nargs="+",
        help=(
             "A set of paths to the controlnet conditioning image be evaluated every `--validation_steps`"
             " and logged to `--report_to`. Should match the number --validation_prompt."
        ),
    )
    parser.add_argument(
        "--num_validation_images",
        type=int,
        default=4,
        help="Number of images to be generated for each `--validation_prompt`",
    )
    parser.add_argument(
        "--validation_steps",
        type=int,
        default=100,
        help=(
            "Run validation every X steps. Validation consists of running the prompt"
            " `args.validation_prompt` multiple times: `args.num_validation_images`"
            " and logging the images."
        ),
    )
    parser.add_argument(
        "--allow_tf32",
        action="store_true",
        help=(
            "Whether or not to allow TF32 on Ampere GPUs. Can be used to speed up training. For more information, see"
            " https://pytorch.org/docs/stable/notes/cuda.html#tensorfloat-32-tf32-on-ampere-devices"
        ),
    )
    parser.add_argument("--local_rank", type=int, default=-1, help="For distributed training: local_rank")
    parser.add_argument(
        "--train_unet_crossattn_kv",
        action="store_true",
        help="Whether or not to fine-tune key, value projection layers in the UNet cross attention.",
    )
    parser.add_argument(
        "--train_unet_crossattn_q_out",
        action="store_true",
        help="Whether or not to fine-tune query, out projection layers in the UNet cross attention.",
    )

    args = parser.parse_args()
    env_local_rank = int(os.environ.get("LOCAL_RANK", -1))
    if env_local_rank != -1 and env_local_rank != args.local_rank:
        args.local_rank = env_local_rank

    return args
    

def main():
    args = parse_args()
    logging_dir = Path(args.output_dir, args.logging_dir)

    accelerator_project_config = ProjectConfiguration(project_dir=args.output_dir, logging_dir=logging_dir)

    accelerator = Accelerator(
        gradient_accumulation_steps=args.gradient_accumulation_steps,
        mixed_precision=args.mixed_precision,
        log_with=args.report_to,
        project_config=accelerator_project_config,
    )

    # If passed along, set the training seed now.
    if args.seed is not None:
        set_seed(args.seed)
    
    if accelerator.is_main_process:
        if args.output_dir is not None:
            os.makedirs(args.output_dir, exist_ok=True)

    # Load scheduler, tokenizer and models.
    noise_scheduler = DDPMScheduler.from_pretrained(args.pretrained_model_name_or_path, subfolder="scheduler")
    #tokenizer = CLIPTokenizer.from_pretrained(args.pretrained_model_name_or_path, subfolder="tokenizer")
    #text_encoder = CLIPTextModel.from_pretrained(args.pretrained_model_name_or_path, subfolder="text_encoder")
    if args.pretrained_vae_model_name_or_path is None:
        vae = AutoencoderKL.from_pretrained(args.pretrained_model_name_or_path, subfolder="vae")
    else:
        vae = AutoencoderKL.from_pretrained(args.pretrained_vae_model_name_or_path, subfolder=None)
    unet = UNet2DConditionModel.from_pretrained(args.pretrained_model_name_or_path, subfolder="unet")
    unet.set_attn_processor(AttnProcessor())
    #image_encoder = CLIPVisionModelWithProjection.from_pretrained(args.image_encoder_path)
    scenegraph_encoder = SceneGraphEncoder(text_encoder_path=args.text_encoder_path, device=accelerator.device)
    # freeze parameters of models to save more memory
    unet.requires_grad_(False)
    vae.requires_grad_(False)
    #text_encoder.requires_grad_(False)
    #image_encoder.requires_grad_(False)
    scenegraph_encoder.text_encoder.requires_grad_(False)
    
    #sg-adapter
    if args.proj_type == 'GCN':
        sg_proj_model = GCNAdapter(
            input_channels=unet.config.cross_attention_dim,
            hidden_channels=unet.config.cross_attention_dim,
            output_channels=unet.config.cross_attention_dim,
        )
    elif args.proj_type == 'GAT':
        sg_proj_model = GATAdapter()
    elif args.proj_type == 'Resampler':
        sg_proj_model = Resampler(
            dim=unet.config.cross_attention_dim,
            depth=4,
            dim_head=64,
            heads=12,
            num_queries=args.num_tokens,
            embedding_dim=unet.config.cross_attention_dim,
            output_dim=unet.config.cross_attention_dim,
            ff_mult=4,
        )
    else:
        raise ValueError("Adapter model not supported!")

    train_kv = args.train_unet_crossattn_kv
    train_q_out = args.train_unet_crossattn_q_out

    attn_procs = {}
    unet_sd = unet.state_dict()
    for name in unet.attn_processors.keys():
        cross_attention_dim = None if name.endswith("attn1.processor") else unet.config.cross_attention_dim
        if name.startswith("mid_block"):
            hidden_size = unet.config.block_out_channels[-1]
        elif name.startswith("up_blocks"):
            block_id = int(name[len("up_blocks.")])
            hidden_size = list(reversed(unet.config.block_out_channels))[block_id]
        elif name.startswith("down_blocks"):
            block_id = int(name[len("down_blocks.")])
            hidden_size = unet.config.block_out_channels[block_id]
        if cross_attention_dim is None:
            #attn_procs[name] = AttnProcessor()
            attn_procs[name] = CustomDiffusionAttnProcessor(
                train_kv=False,
                train_q_out=False,
                hidden_size=hidden_size,
                cross_attention_dim=cross_attention_dim,
            )
        else:
            layer_name = name.split(".processor")[0]
            #weights = {
            #    "to_k_ip.weight": unet_sd[layer_name + ".to_k.weight"],
            #    "to_v_ip.weight": unet_sd[layer_name + ".to_v.weight"],
            #}
            if train_kv:
                weights = {
                    "to_k_custom_diffusion.weight": unet_sd[layer_name + ".to_k.weight"],
                    "to_v_custom_diffusion.weight": unet_sd[layer_name + ".to_v.weight"],
                }
            if train_q_out:
                weights["to_q_custom_diffusion.weight"] = unet_sd[layer_name + ".to_q.weight"]
                weights["to_out_custom_diffusion.0.weight"] = unet_sd[layer_name + ".to_out.0.weight"]
                weights["to_out_custom_diffusion.0.bias"] = unet_sd[layer_name + ".to_out.0.bias"]
            #attn_procs[name] = IPAttnProcessor(hidden_size=hidden_size, cross_attention_dim=cross_attention_dim, num_tokens=args.num_tokens)
            attn_procs[name] = CustomDiffusionAttnProcessor(
                train_kv=train_kv,
                train_q_out=train_q_out,
                hidden_size=hidden_size,
                cross_attention_dim=cross_attention_dim,
            ).to(unet.device)
            if train_kv or train_q_out:
                attn_procs[name].load_state_dict(weights)
    del unet_sd
    unet.set_attn_processor(attn_procs)
    adapter_modules = torch.nn.ModuleList(unet.attn_processors.values())
    #custom_diffusion_layers = AttnProcsLayers(unet.attn_processors)
    #accelerator.register_for_checkpointing(custom_diffusion_layers)
    
    sg_adapter = TrainAdapterHelper(
        unet=unet,
        sg_proj_model=sg_proj_model,
        adapter_modules=adapter_modules,
        ckpt_path=args.pretrained_sg_adapter_path
    )

    # For mixed precision training we cast the text_encoder and vae weights to half-precision
    # as these models are only used for inference, keeping weights in full precision is not required.
    ########
    weight_dtype = torch.float32
    if accelerator.mixed_precision == "fp16":
        weight_dtype = torch.float16
    elif accelerator.mixed_precision == "bf16":
        weight_dtype = torch.bfloat16
    #unet.to(accelerator.device, dtype=weight_dtype)
    vae.to(accelerator.device, dtype=weight_dtype)
    #text_encoder.to(accelerator.device, dtype=weight_dtype)
    #image_encoder.to(accelerator.device, dtype=weight_dtype)
    scenegraph_encoder.text_encoder.to(accelerator.device, dtype=weight_dtype)
    ########

    # Check that all trainable models are in full precision
    low_precision_error_string = (
        " Please make sure to always have all model weights in full float32 precision when starting training - even if"
        " doing mixed precision training, copy of the weights should still be float32."
    )

    if accelerator.unwrap_model(sg_adapter).unet.dtype != torch.float32:
        raise ValueError(
            f"UNet loaded as datatype {accelerator.unwrap_model(sg_adapter).unet.dtype}. {low_precision_error_string}"
        )

    # Enable TF32 for faster training on Ampere GPUs,
    # cf https://pytorch.org/docs/stable/notes/cuda.html#tensorfloat-32-tf32-on-ampere-devices
    if args.allow_tf32:
        torch.backends.cuda.matmul.allow_tf32 = True

    if args.scale_lr:
        args.learning_rate = (
            args.learning_rate * args.gradient_accumulation_steps * args.train_batch_size * accelerator.num_processes
        )
    
    # optimizer
    params_to_opt = itertools.chain(sg_adapter.sg_proj_model.parameters(),  sg_adapter.adapter_modules.parameters())
    optimizer = torch.optim.AdamW(
        params_to_opt,
        lr=args.learning_rate,
        betas=(args.adam_beta1, args.adam_beta2),
        weight_decay=args.weight_decay,
        eps=args.adam_epsilon,
    )

    # dataloader
    train_dataset = CustomDataset(
        image_dir=args.images_root_path,
        data_dir=args.data_root_path,
        transform=get_transforms_rand_crop()
    )
    train_dataloader = torch.utils.data.DataLoader(
        train_dataset,
        shuffle=True,
        collate_fn=collate_fn,
        batch_size=args.train_batch_size,
        num_workers=args.dataloader_num_workers,
    )

    # Scheduler and math around the number of training steps.
    overrode_max_train_steps = False
    num_update_steps_per_epoch = math.ceil(len(train_dataloader) / args.gradient_accumulation_steps)
    if args.max_train_steps is None:
        args.max_train_steps = args.num_train_epochs * num_update_steps_per_epoch
        overrode_max_train_steps = True

    lr_scheduler = get_scheduler(
        args.lr_scheduler,
        optimizer=optimizer,
        num_warmup_steps=args.lr_warmup_steps * accelerator.num_processes,
        num_training_steps=args.max_train_steps * accelerator.num_processes,
        num_cycles=args.lr_num_cycles,
        power=args.lr_power,
    )

    # Prepare everything with our `accelerator`.
    sg_adapter, optimizer, train_dataloader, lr_scheduler = accelerator.prepare(
        sg_adapter, optimizer, train_dataloader, lr_scheduler
    )

    # We need to recalculate our total training steps as the size of the training dataloader may have changed.
    num_update_steps_per_epoch = math.ceil(len(train_dataloader) / args.gradient_accumulation_steps)
    if overrode_max_train_steps:
        args.max_train_steps = args.num_train_epochs * num_update_steps_per_epoch
    # Afterwards we recalculate our number of training epochs
    args.num_train_epochs = math.ceil(args.max_train_steps / num_update_steps_per_epoch)

    # We need to initialize the trackers we use, and also store our configuration.
    # The trackers initialize automatically on the main process.
    if accelerator.is_main_process:
        if args.tracker_log_dir is not None:
            os.makedirs(args.tracker_log_dir, exist_ok=True)
        tracker_config = dict(vars(args))
        init_kwargs = None
        if args.tracker_run_name is not None:
            init_kwargs = {}
            init_kwargs["name"] = args.tracker_run_name.replace("{TIME}", datetime.now().strftime("%d/%m/%Y-%H:%M:%S"))
        if args.report_to == "wandb":
            init_kwargs = {"wandb": {}}
            init_kwargs["wandb"]["dir"] = args.tracker_log_dir
            #init_kwargs["wandb"]["project"] = args.tracker_project_name
            init_kwargs["wandb"]["name"] = args.tracker_run_name.replace("{TIME}", datetime.now().strftime("%d/%m/%Y-%H:%M:%S"))
        # tensorboard cannot handle list types for config
        tracker_config.pop("validation_prompt")
        tracker_config.pop("validation_image")
        accelerator.init_trackers(args.tracker_project_name, config=tracker_config, init_kwargs=init_kwargs)

    # Train!
    total_batch_size = args.train_batch_size * accelerator.num_processes * args.gradient_accumulation_steps
    #max_train_steps = args.num_train_epochs * len(train_dataloader.dataset)
    empty_node_embed = scenegraph_encoder.get_empty_node_embed().squeeze()

    logger.info("***** Running training *****")
    logger.info(f"  Num examples = {len(train_dataset)}")
    logger.info(f"  Num batches each epoch = {len(train_dataloader)}")
    logger.info(f"  Num Epochs = {args.num_train_epochs}")
    logger.info(f"  Instantaneous batch size per device = {args.train_batch_size}")
    logger.info(f"  Total train batch size (w. parallel, distributed & accumulation) = {total_batch_size}")
    logger.info(f"  Gradient Accumulation steps = {args.gradient_accumulation_steps}")
    logger.info(f"  Total optimization steps = {args.max_train_steps}")
    global_step = 0
    first_epoch = 0

    # Potentially load in the weights and states from a previous save
    if args.resume_from_checkpoint:
        if args.resume_from_checkpoint != "latest":
            path = os.path.basename(args.resume_from_checkpoint)
        else:
            # Get the most recent checkpoint
            dirs = os.listdir(args.output_dir)
            dirs = [d for d in dirs if d.startswith("checkpoint")]
            dirs = sorted(dirs, key=lambda x: int(x.split("-")[1]))
            path = dirs[-1] if len(dirs) > 0 else None

        if path is None:
            accelerator.print(
                f"Checkpoint '{args.resume_from_checkpoint}' does not exist. Starting a new training run."
            )
            args.resume_from_checkpoint = None
            initial_global_step = 0
        else:
            accelerator.print(f"Resuming from checkpoint {path}")
            accelerator.load_state(os.path.join(args.output_dir, path))
            global_step = int(path.split("-")[1])

            initial_global_step = global_step
            first_epoch = global_step // num_update_steps_per_epoch
    else:
        initial_global_step = 0

    progress_bar = tqdm(
        range(0, args.max_train_steps),
        initial=initial_global_step,
        desc="Steps",
        # Only show the progress bar once on each machine.
        disable=not accelerator.is_local_main_process,
    )

    image_logs = log_validation(
        vae,
        # text_encoder,
        # tokenizer,
        sg_adapter.unet,
        # image_encoder,
        scenegraph_encoder,
        sg_adapter.sg_proj_model,
        args,
        accelerator,
        weight_dtype,
        global_step,
    )

    for epoch in range(first_epoch, args.num_train_epochs):
        #begin = time.perf_counter()
        for step, batch in enumerate(train_dataloader):
            #load_data_time = time.perf_counter() - begin
            with accelerator.accumulate(sg_adapter):
                # Convert images to latent space
                with torch.no_grad():
                    latents = vae.encode(batch["images"].to(accelerator.device, dtype=weight_dtype)).latent_dist.sample()
                    latents = latents * vae.config.scaling_factor

                # Sample noise that we'll add to the latents
                noise = torch.randn_like(latents)
                bsz = latents.shape[0]
                # Sample a random timestep for each image
                timesteps = torch.randint(0, noise_scheduler.num_train_timesteps, (bsz,), device=latents.device)
                timesteps = timesteps.long()

                # Add noise to the latents according to the noise magnitude at each timestep
                # (this is the forward diffusion process)
                noisy_latents = noise_scheduler.add_noise(latents, noise, timesteps)

                with torch.no_grad():
                    scene_graphs_embeds, _ = scenegraph_encoder(
                        batch["scene_graphs"],
                        padding=isinstance(sg_proj_model, Resampler),
                        do_classifier_free_guidance=False
                    )
                edge_index = batch["scene_graphs"].adj_list.to(accelerator.device)

                # Non usiamo testo
                #with torch.no_grad():
                #    encoder_hidden_states = text_encoder(batch["text_input_ids"].to(accelerator.device))[0]
                ###############
                
                noise_pred = sg_adapter(
                    noisy_latents,
                    timesteps,
                    scene_graphs_embeds,
                    edge_index,
                    batch["node_counts"],
                    empty_node_embed
                )

                loss = F.mse_loss(noise_pred.float(), noise.float(), reduction="mean")
            
                # Gather the losses across all processes for logging (if we use distributed training).
                avg_loss = accelerator.gather(loss.repeat(args.train_batch_size)).mean().item()
                
                # Backpropagate
                accelerator.backward(loss)
                if accelerator.sync_gradients:
                    params_to_clip = params_to_opt
                    accelerator.clip_grad_norm_(params_to_clip, args.max_grad_norm)
                optimizer.step()
                lr_scheduler.step()
                optimizer.zero_grad(set_to_none=args.set_grads_to_none)

            # Checks if the accelerator has performed an optimization step behind the scenes
            if accelerator.sync_gradients:
                progress_bar.update(1)
                global_step += 1

                if accelerator.is_main_process:
                    if global_step % args.checkpointing_steps == 0:
                        # _before_ saving state, check if this save would set us over the `checkpoints_total_limit`
                        if args.checkpoints_total_limit is not None:
                            checkpoints = os.listdir(args.output_dir)
                            checkpoints = [d for d in checkpoints if d.startswith("checkpoint")]
                            checkpoints = sorted(checkpoints, key=lambda x: int(x.split("-")[1]))

                            # before we save the new checkpoint, we need to have
                            # at _most_ `checkpoints_total_limit - 1` checkpoints
                            if len(checkpoints) >= args.checkpoints_total_limit:
                                num_to_remove = len(checkpoints) - args.checkpoints_total_limit + 1
                                removing_checkpoints = checkpoints[0:num_to_remove]

                                logger.info(
                                    f"{len(checkpoints)} checkpoints already exist, removing {len(removing_checkpoints)} checkpoints"
                                )
                                logger.info(f"removing checkpoints: {', '.join(removing_checkpoints)}")

                                for removing_checkpoint in removing_checkpoints:
                                    removing_checkpoint = os.path.join(args.output_dir, removing_checkpoint)
                                    shutil.rmtree(removing_checkpoint)

                        save_path = os.path.join(args.output_dir, f"checkpoint-{global_step}")
                        accelerator.save_state(save_path, safe_serialization=True)
                        logger.info(f"Saved state to {save_path}")

                    if args.validation_prompt is not None and global_step % args.validation_steps == 0:
                        image_logs = log_validation(
                            vae,
                            #text_encoder,
                            #tokenizer,
                            sg_adapter.unet,
                            #image_encoder,
                            scenegraph_encoder,
                            sg_adapter.sg_proj_model,
                            args,
                            accelerator,
                            weight_dtype,
                            global_step,
                        )
            
            #global_step += 1
            
            #if global_step % args.save_steps == 0:
            #    save_path = os.path.join(args.output_dir, f"checkpoint-{global_step}")
            #    accelerator.save_state(save_path)

            logs = {"loss": loss.detach().item(), "avg_loss": avg_loss, "lr": lr_scheduler.get_last_lr()[0]}
            progress_bar.set_postfix(**logs)
            accelerator.log(logs, step=global_step)

            if global_step >= args.max_train_steps:
                break
            
            #begin = time.perf_counter()

    # Create the pipeline using the trained modules and save it.
    accelerator.wait_for_everyone()
    if accelerator.is_main_process:
        sg_adapter = accelerator.unwrap_model(sg_adapter)
        sg_adapter.save_pretrained(args.output_dir)

    accelerator.end_training()


def log_validation(vae, unet, sg_encoder, sg_proj_model, args, accelerator, weight_dtype, step):
    logger.info("Running validation... ")

    pipeline = StableDiffusionPipeline.from_pretrained(
        args.pretrained_model_name_or_path,
        vae=vae,
        text_encoder=None,
        tokenizer=None,
        unet=unet,
        safety_checker=None,
        torch_dtype=weight_dtype,
    )
    pipeline.scheduler = DPMSolverMultistepScheduler.from_config(
        pipeline.scheduler.config,
        algorithm_type="sde-dpmsolver++",
        use_karras_sigmas='true'
    )
    pipeline = pipeline.to(accelerator.device)
    pipeline.set_progress_bar_config(disable=True)

    #if args.seed is None:
    #    generator = None
    #else:
    #    generator = torch.Generator(device=accelerator.device).manual_seed(args.seed)

    if len(args.validation_image) == len(args.validation_prompt):
        validation_images = args.validation_image
        validation_prompts = args.validation_prompt
    else:
        raise ValueError(
            "number of `args.validation_image` and `args.validation_prompt` should be checked in `parse_args`"
        )

    sg_model = SGDiffuser(
        sd_pipe=pipeline,
        sg_encoder=sg_encoder,
        text_encoder_path=None,
        tokenizer=None,
        proj_type=sg_proj_model,
        ip_ckpt=None,
        use_text=False,
        device=accelerator.device
    )

    image_logs = []

    for validation_prompt, validation_image in zip(validation_prompts, validation_images):
        validation_image = Image.open(validation_image).convert("RGB")
        with open(validation_prompt, 'r') as f:
            data = json.load(f)
        nodes = {**data["EL_dict"], **data["rel_dict"]}
        sorted_nodes = sorted(nodes, key=nodes.get)
        edge_index = torch.tensor(data["adj_list"]).t().contiguous()
        validation_prompt = SceneGraph(
            nodes=sorted_nodes,
            adj_list=edge_index,
            node_count=len(sorted_nodes),
            edge_count=edge_index.shape[1]
        )

        images = []

        for k in range(args.num_validation_images):
            with torch.autocast("cuda"):
                image = sg_model.generate(
                    prompt=validation_prompt,
                    num_inference_steps=30,
                    guidance_scale=0,
                    height=512,
                    width=512,
                    seed=args.seed + k
                )[0]

            images.append(image)

        image_logs.append(
            {"validation_image": validation_image, "images": images, "validation_prompt": data['caption']}
        )

    for tracker in accelerator.trackers:
        if tracker.name == "tensorboard":
            for log in image_logs:
                images = log["images"]
                validation_prompt = log["validation_prompt"]
                validation_image = log["validation_image"]

                formatted_images = []

                formatted_images.append(np.asarray(validation_image))

                for image in images:
                    formatted_images.append(np.asarray(image))

                formatted_images = np.stack(formatted_images)

                tracker.writer.add_images(validation_prompt, formatted_images, step, dataformats="NHWC")
        elif tracker.name == "wandb":
            formatted_images = []

            for log in image_logs:
                images = log["images"]
                validation_prompt = log["validation_prompt"]
                validation_image = log["validation_image"]

                formatted_images.append(wandb.Image(validation_image, caption="Ground truth", file_type="jpg"))

                for image in images:
                    formatted_images.append(wandb.Image(image, caption=validation_prompt, file_type="jpg"))
                    #image = wandb.Image(image, caption=validation_prompt)
                    #formatted_images.append(image)

            tracker.log({"validation": formatted_images})
        else:
            logger.warn(f"image logging not implemented for {tracker.name}")

        return image_logs


if __name__ == "__main__":
    main()
